//mini net :
/* node 
*/
class snn_node{
       public:
double weights[];//weights to the previous layer
double weight_adjustments[];//the "learning" that will occur
double bias,bias_adjustment;//the bias unit (a weight with a node of value 1.0 essentially)
double error;//we receive the error from the following layers here
double output;//the output value of the node
       snn_node(void){reset();}
      ~snn_node(void){reset();}
  void reset(){
       ArrayFree(weights);
       ArrayFree(weight_adjustments);
       bias=0.0;
       bias_adjustment=0.0;
       error=0.0;
       output=0.0;
       }
  void setup(int nodes_of_previous_layer,
             bool randomize_weights=true,
             double weights_min=-1.0,
             double weights_max=1.0){
       //we have the same number of weights as nodes of the previous layer
       ArrayResize(weights,nodes_of_previous_layer,0);
       ArrayResize(weight_adjustments,nodes_of_previous_layer,0);
       ArrayFill(weight_adjustments,0,nodes_of_previous_layer,0.0);
       //if weights randomize 
         if(randomize_weights){
         init_weights(weights_min,weights_max);
         }else{
         ArrayFill(weights,0,nodes_of_previous_layer,1.0);
         bias=1.0;
         }
       }
       //initialize weights
  void init_weights(double min,double max){
       double range=max-min;
       for(int i=0;i<ArraySize(weights);i++){
          weights[i]=(((double)MathRand())/32767.0)*range+min;
          }
       bias=(((double)MathRand())/32767.0)*range+min;
       }
       /* now the calculation
          for this node to be calculated (feed forward)
          we need the nodes of the previous layer
       */
  void feed_forward(snn_node &previous_layer_nodes[]){
       output=0.0;
       error=0.0;
       //we sum the weights times the previous layer node values 
         for(int i=0;i<ArraySize(previous_layer_nodes);i++){
            output+=previous_layer_nodes[i].output*weights[i];
            }
         //we add the bias
           output+=bias;
         //we send this through the activation function
           output=activationSigmoid(output);     
       }
       /* the back propagation
          we have received an error from the loss function 
          (if an output node)
          or from following layers
       */
  void back_propagation(bool propagate_error,//if true we will send the error to the previous layer , this will be false if the previous layer is the input layer
                        snn_node &previous_layer_nodes[]){//the previous layer's nodes to receive the error
       /* imagine the error received is after the activation function of this node
          so , we need to bring it back to distribute it 
          we multiply it by the derivative of the activation function
       */
       error*=derivativeSigmoid(output);
       //if we propagate the error to the previous layer 
         if(propagate_error){
           for(int i=0;i<ArraySize(previous_layer_nodes);i++){ 
              /*
              the error to each previous node is the error prior activation here x the weight to that node
              why do we add it ? because there may be more nodes connecting back to that previous layer node so
              we want the total error , we also add to the accumulation of the error but that is not reset on each 
              forward pass
              we use the weight before the "learning" not the new one
              */
              previous_layer_nodes[i].error+=weights[i]*error;
              }
           }
       //and add to the adjustments needed 
         for(int i=0;i<ArraySize(weights);i++){
            weight_adjustments[i]+=error*previous_layer_nodes[i].output;
            }
         //adjust the bias 
           bias_adjustment+=error;
       }
  void adjust(double learning_rate){
       for(int i=0;i<ArraySize(weights);i++){
          weights[i]-=learning_rate*weight_adjustments[i];
          }
       bias-=learning_rate*bias_adjustment;
       reset_weight_adjustments();
       }
  void reset_weight_adjustments(){
       for(int i=0;i<ArraySize(weight_adjustments);i++){
          weight_adjustments[i]=0.0;
          }
       bias_adjustment=0.0;
       }
};
/*
a layer
*/
class snn_layer{
         public:
snn_node nodes[];
         snn_layer(void){reset();}
        ~snn_layer(void){reset();}
    void reset(){
         ArrayFree(nodes);
         }
    void setup(int nodes_total){
         ArrayResize(nodes,nodes_total,0);
         }
    void setup(int nodes_total,
               int previous_layer_nodes_total,
               bool randomize_weights=true,
               double weights_min=-1.0,
               double weights_max=1.0){
         //we send the previous layer so we can setup the nodes as well
           ArrayResize(nodes,nodes_total,0);
           for(int i=0;i<ArraySize(nodes);i++){
              nodes[i].setup(previous_layer_nodes_total,randomize_weights,weights_min,weights_max);
              }
         }
         //fill if input layer
    void fill_as_input_layer(double &inputs[]){
         for(int i=0;i<ArraySize(nodes);i++){
            nodes[i].output=inputs[i];//we place the values straight to the output
            }
         }
         //feed forward / calculate
    void feed_forward(snn_layer &previous_layer){
         for(int i=0;i<ArraySize(nodes);i++){
            nodes[i].feed_forward(previous_layer.nodes);
            }
         }
         //back propagation
    void back_propagation(bool propagate_error,
                          snn_layer &previous_layer){
         for(int i=0;i<ArraySize(nodes);i++){
            nodes[i].back_propagation(propagate_error,previous_layer.nodes);
            }
         }
         /* and
            if this is the output layer we wil
            need to calculate the error via direct comparison
            with the correct output! so 
         */
  double output_layer_calculate_loss(double &correct_outputs[]){
         //the correct outputs are the same # as the nodes of the output layer
           //the loss function we'll use is the quadratic 
           //its :  where i is the total output nodes
             //this is the value we'll return to the network , not the backpropagation value
               //you could use this to control the learning rate for instance
                 double loss=0.0;
                 for(int i=0;i<ArraySize(correct_outputs);i++){
                 loss+=MathPow((nodes[i].output-correct_outputs[i]),2.0);
                 }
                 loss/=2.0;
         //and the error per output node is the derivative of 1/2(f-c)^2
         //which is : f-c
           for(int i=0;i<ArraySize(nodes);i++){
              //so the error post activation for output node i is : 
                nodes[i].error=nodes[i].output-correct_outputs[i];
              }
         return(loss);
         } 
    void adjust(double learning_rate){
         for(int i=0;i<ArraySize(nodes);i++){
            nodes[i].adjust(learning_rate);
            }
         }
};

class snn_net{
           public:
 snn_layer layers[];
       int samples_calculated;//for the batch
    double total_loss;
           snn_net(void){reset();}
          ~snn_net(void){reset();}
      void reset(){
           ArrayFree(layers);
           samples_calculated=0;
           total_loss=0.0;
           }
snn_layer* add_layer(){
           int ns=ArraySize(layers)+1;
           ArrayResize(layers,ns,0);
           return(GetPointer(layers[ns-1]));
           } 
      void feed_forward(double &inputs[]){
           //fill the input layer 
             layers[0].fill_as_input_layer(inputs);
           //and for each following layer 
             for(int i=1;i<ArraySize(layers);i++){
                layers[i].feed_forward(layers[i-1]);
                }
                
           }
      void calculate_loss(double &sample_correct_outputs[]){
           samples_calculated++;
           //calculate the loss at the output layer and also add to the total loss
           total_loss+=layers[ArraySize(layers)-1].output_layer_calculate_loss(sample_correct_outputs);
           }
      void back_propagation(){
           //we start from the output layer - for which we must've calculated the error first
             for(int i=ArraySize(layers)-1;i>=1;i--){
                //and we send a back prop call but if the back prop is from layer[1] to layer[0] we
                //do not send the error back (because its the input layer)
                  layers[i].back_propagation(((bool)(i>1)),layers[i-1]);
                }
           }
      void adjust(double learning_rate){
           for(int i=1;i<ArraySize(layers);i++){
              layers[i].adjust(learning_rate);
              }
           }
    double get_max_loss_per_sample(){
           //the max value per output node of the output layer is 1.0 and the min 0.0
           //this means the maximum loss per sample would have a deviation of -1.0 or 1.0
           //we plug -1 in the quadratic loss function 
           //1/2(S(i)(forecast-correct)^2) , if all output nodes are wrong for one sample
             //but the forecast-correct squared of -1 or 1 will always output 1 so 
             //the max loss per sample is (1*output_nodes)/2
             double max_loss_per_sample=((double)ArraySize(layers[ArraySize(layers)-1].nodes))/2.0;
           return(max_loss_per_sample);
           }
    double get_current_loss_per_sample(){
           double current_loss_per_sample=total_loss/((double)samples_calculated);
           return(current_loss_per_sample);
           }
    double get_output(int ix){
           return(layers[ArraySize(layers)-1].nodes[ix].output);
           }
      void reset_loss(){
           total_loss=0.0;
           samples_calculated=0;
           }
};
//various
/* first the sigmoid activation function */
double activationSigmoid(double x){
return(1/(1+MathExp(-1.0*x)));
}
// then the derivative of sigmoid (we receive the activation value )
/*
so this is essentially , if x=1.0 derivativeSigmoid(activationSigmoid(1.0)); 
*/
double derivativeSigmoid(double activation_value){
return(activation_value*(1-activation_value));
}

